@@ -15,5 +15,5 @@
 from .axes_scan import broadcast
 from .frozen_dict import FrozenDict, freeze, unfreeze
 from .tracers import current_trace, trace_level, check_trace_level
-from .scope import Scope, Array, apply, init, bind
+from .scope import Scope, Array, DenyList, apply, init, bind
 from .lift import scan, vmap, jit
