@@ -15,13 +15,9 @@
 """Linear modules."""
 
 from collections.abc import Iterable  # pylint: disable=g-importing-member
-
-from flax.nn import initializers
-
-from flax.core import Scope
-
 from flax import struct
-
+from flax.core import Scope
+from flax.deprecated.nn import initializers
 from jax import lax
 
 import jax.numpy as jnp
