@@ -16,20 +16,22 @@
 
 import jax
 
+from .. import errors
+
 
 def current_trace():
   """Returns the innermost Jax tracer."""
   return jax.core.find_top_trace(())
 
 
-def trace_level(master):
+def trace_level(main):
   """Returns the level of the trace of -infinity if it is None."""
-  if master:
-    return master.level
+  if main:
+    return main.level
   return float('-inf')
 
 
 def check_trace_level(base_level):
   level = trace_level(current_trace())
   if level != base_level:
-    raise ValueError('Jax transforms and modules cannot be mixed.')
+    raise errors.JaxTransformError()
