@@ -42,9 +42,10 @@ def pool(inputs, init, reduce_fn, window_shape, strides, padding):
     The output of the reduction for each window slice.
   """
   strides = strides or (1,) * len(window_shape)
+  assert len(window_shape) == len(strides), (
+      f"len({window_shape}) == len({strides})")
   strides = (1,) + strides + (1,)
   dims = (1,) + window_shape + (1,)
-  assert len(dims) == len(strides)
 
   is_single_input = False
   if inputs.ndim == len(dims) - 1:
@@ -53,7 +54,7 @@ def pool(inputs, init, reduce_fn, window_shape, strides, padding):
     inputs = inputs[None]
     is_single_input = True
 
-  assert inputs.ndim == len(dims)
+  assert inputs.ndim == len(dims), f"len({inputs.shape}) != len({dims})"
   if not isinstance(padding, str):
     padding = tuple(map(tuple, padding))
     assert(len(padding) == len(window_shape)), (
@@ -103,3 +104,21 @@ def max_pool(inputs, window_shape, strides=None, padding="VALID"):
   """
   y = pool(inputs, -jnp.inf, lax.max, window_shape, strides, padding)
   return y
+
+
+def min_pool(inputs, window_shape, strides=None, padding="VALID"):
+  """Pools the input by taking the minimum of a window slice.
+
+  Args:
+    inputs: Input data with dimensions (batch, window dims..., features).
+    window_shape: A shape tuple defining the window to reduce over.
+    strides: A sequence of `n` integers, representing the inter-window strides
+        (default: `(1, ..., 1)`).
+    padding: Either the string `'SAME'`, the string `'VALID'`, or a sequence of
+      `n` `(low, high)` integer pairs that give the padding to apply before and
+      after each spatial dimension (default: `'VALID'`).
+
+  Returns:
+    The minimum for each window slice.
+  """
+  return pool(inputs, jnp.inf, lax.min, window_shape, strides, padding)
