@@ -12,11 +12,21 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""Flax Optimizer api."""
+"""Flax Optimizer api.
+
+Note that with `FLIP #1009`_ the optimizers in ``flax.optim`` were **effectively
+deprecated** in favor of Optax_ optimizers. By now, ``optax`` should support all
+of the original features from ``flax.optim`` (otherwise please create a Github
+issue on ``optax``).
+
+.. _FLIP #1009: https://github.com/google/flax/blob/main/docs/flip/1009-optimizer-api.md
+.. _Optax: https://github.com/deepmind/optax
+"""
 
 # pylint: disable=g-multiple-import
 # re-export commonly used modules and functions
 from .adadelta import Adadelta
+from .adabelief import AdaBelief
 from .adafactor import Adafactor
 from .adagrad import Adagrad
 from .adam import Adam
@@ -31,6 +41,7 @@ from .weight_norm import WeightNorm
 
 __all__ = [
     "Adam",
+    "AdaBelief",
     "Adadelta",
     "Adafactor",
     "Adagrad",
